"""main.py."""
import logging

from fastapi import FastAPI, Request, status
from fastapi.responses import JSONResponse
from strangeworks_core.errors.error import StrangeworksError

from .middleware import RequestContextMiddleware
from .routes import router

logging.basicConfig(level=logging.INFO)
gql_logger = logging.getLogger("gql.transport.requests")
gql_logger.setLevel(logging.DEBUG)

app = FastAPI()


@app.exception_handler(StrangeworksError)
async def http_exception_handler(
    request: Request, exc: StrangeworksError
) -> JSONResponse:
    """Exception handler for StrangeworksError exceptions.

    More info on how this works here: https://tinyurl.com/2hfr9fwj

    Parameters
    ----------
    requst: Request
        request object.
    exc: StrangeworksError
        StrangeworksError exception.

    Return
    ------
    :JSONResponse
        The exception represented as a JSONResponse object.
    """
    logging.exception(exc)
    return JSONResponse(status_code=400, content={"error": exc.message})


# add routers
app.add_middleware(RequestContextMiddleware)

app.include_router(router)


@app.get("/", status_code=status.HTTP_200_OK)
async def health_check():
    return {"status": "ok"}



"""SubmitRequest type."""
from dataclasses import dataclass

from pydantic import BaseModel
from strangeworks_core.types.job import Job


class SubmitRequest(BaseModel):
    """Submit Request type.

    Used for making requests to start a job. Allows caller to
    create the job entry on the platform first.
    """

    callable_params: dict | None = None


class JobRequest(BaseModel):
    """Job Request model.

    Used for making requests for existing jobs.

    Parameters
    ----------
    slug: str
        job identifier.
    """

    slug: str


@dataclass
class JobResponse:
    """Submit Response type.

    Purpose is to allow the response to expand if needed beyond platform job info in
    the future.
    """

    job: Job



"""callable.py."""

from fastapi import APIRouter, Request

from .handler import execute_request
from .middleware import RequestParams
from .types import SubmitRequest

router: APIRouter = APIRouter(prefix="/jobs", tags=["Callables"])


@router.post("/submit")
async def run_callable(request: Request, submit_request: SubmitRequest):
    params: RequestParams = request.state.params
    return execute_request(
        ctx=params.ctx,
        app_fn=params.callable_fn,
        estimator_fn=params.estimator_fn,
        cost_fn=params.calculator_fn,
        submit_request=submit_request,
    )



"""common.py."""

import json  # noqa E402
from dataclasses import dataclass

from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from sw_product_lib.platform.gql import ProductAPI
from sw_product_lib.service import DEFAULT_PLATFORM_BASE_URL, RequestContext

from . import AppFunction, CallableImplenmetation, CostCalculator, CostEstimator, _cfg

ignore_paths = ["/docs", "/openapi.json", "/"]


class RequestContextMiddleware(BaseHTTPMiddleware):
    """Add RequestContext to the request."""

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)

    async def dispatch(self, request, call_next):  # noqa
        if request.url.path in ignore_paths:
            return await call_next(request)

        request.state.params = RequestParams.from_request(request)
        return await call_next(request)


@dataclass
class RequestParams:
    """Class for items needed to handle a request."""

    ctx: RequestContext
    callable_fn: AppFunction
    estimator_fn: AppFunction
    calculator_fn: AppFunction

    @classmethod
    def from_request(cls, request: Request) -> "RequestParams":
        ctx = RequestContext.from_request(request=request)
        ctx.api = ProductAPI(
            api_key=_cfg.get("api_key", profile="platform"),
            base_url=_cfg.get("api_url", "platform") or DEFAULT_PLATFORM_BASE_URL,
        )
        return cls(
            ctx=ctx,
            callable_fn=CallableImplenmetation,
            estimator_fn=CostEstimator,
            calculator_fn=CostCalculator,
        )



"""handler.py."""

import logging

from strangeworks_core.errors.error import StrangeworksError
from sw_product_lib import service
from sw_product_lib.service import RequestContext
from sw_product_lib.types.job import Job, JobStatus

from . import AppFunction
from .types import SubmitRequest


def execute_request(
    ctx: RequestContext,
    app_fn: AppFunction,
    estimator_fn: AppFunction,
    cost_fn: AppFunction,
    submit_request: SubmitRequest,
) -> Job:
    """Execute Callable Request."""
    # Even if the job submission fails, there should be a record of it on the platform.
    sw_job: Job = service.create_job(ctx)

    cost_estimate = estimator_fn(submit_request)
    try:
        if not service.request_job_clearance(ctx=ctx, amount=cost_estimate):
            # did not get clearance. update job status to failed and raise error.
            service.update_job(
                ctx=ctx,
                job_slug=sw_job.slug,
                status=JobStatus.FAILED,
            )
            raise StrangeworksError(
                message=f"Job clearance denied for this resource {ctx.resource_slug}."
            )
    except Exception as err:
        service.update_job(
            ctx=ctx,
            job_slug=sw_job.slug,
            status=JobStatus.FAILED,
        )
        raise StrangeworksError(
            message=f"Request to platform for job clearance failed {ctx.resource_slug}."
        ) from err

    try:
        logging.info(f"executing job request (job slug: {sw_job.slug})")
        result = app_fn(**submit_request.callable_params)
        logging.info(f"uploading job result (job slug: {sw_job.slug})")
        service.upload_job_artifact(
            result,
            ctx=ctx,
            job_slug=sw_job.slug,
            file_name="result.json",
        )
        logging.info(f"updating job status to COMPLETED (job slug: {sw_job.slug})")
        service.update_job(
            ctx=ctx,
            job_slug=sw_job.slug,
            status=JobStatus.COMPLETED,
        )
        cost = cost_fn(result)
        logging.info(
            f"creating billing transaction of ${cost} (job slug: {sw_job.slug})"
        )
        try:
            service.create_billing_transaction(
                ctx,
                job_slug=sw_job.slug,
                amount=cost,
                description=f"cost for {sw_job.slug}",
            )
        except Exception as err:
            logging.error(
                f"unable to create billing transaction of {cost} (job slug: {sw_job.slug})"  # noqa
            )
            raise StrangeworksError(message=err) from err
        logging.info(f"completed execution of job request (job slug: {sw_job.slug})")
    except Exception as err:
        service.update_job(
            ctx=ctx,
            job_slug=sw_job.slug,
            status=JobStatus.FAILED,
        )
        raise StrangeworksError(message=err) from err

    return sw_job



"""__init__.py."""

import inspect
from importlib.util import find_spec, module_from_spec
from typing import Callable, ParamSpec, TypeVar

from strangeworks_core.config.config import Config

_cfg = Config()

# better way to provide a parameter spec for a Callable parameter types
# see https://docs.python.org/3/library/typing.html#typing.ParamSpec
T = TypeVar("T")  # can be anything
P = ParamSpec("P")
AppFunction = Callable[P, T]


def _get_callable(module: str, callable_name: str) -> AppFunction:
    """Loads Callable.

    Load the Callable/Function specified by callable_name defined in module.
    Imitates `from module import callable_name` at runtime.

    Parameters
    ----------
    module: str
        fully qualified name of the module where the callable is implemented.
    callable_name: str
        name of the function.

    Returns
    -------
    :AppFunction
        Callable to be used as the implementation of the service.
    """
    mod_spec = find_spec(module)
    if mod_spec is None:
        raise ValueError(f"unable to find the {module} module.")
    mod = module_from_spec(mod_spec)
    mod_spec.loader.exec_module(mod)

    _functions = inspect.getmembers(mod, inspect.isfunction)

    for name, _fn in _functions:
        if name == callable_name:
            return _fn
    return None


# set up the function that will be called when the service is invoked.

_CALLABLE_NAME = _cfg.get("name", profile="callable_svc")
_CALLABLE_MODULE = _cfg.get("module", profile="callable_svc")
CallableImplenmetation: AppFunction = _get_callable(
    module=_CALLABLE_MODULE, callable_name=_CALLABLE_NAME
)


# default functions
def _default_cost_estimator(*args, **kwargs) -> float:
    return -1.0


def _default_cost_calculator(*args, **kwargs) -> float:
    return 1.0


# set up cost estimator function. default to a function that returns 0.
_COST_ESIMATOR_NAME = _cfg.get("cost_estimator", profile="callable_svc")
_COST_ESIMATOR_MODULE = (
    _cfg.get("cost_estimator_module", profile="callable_svc") or _CALLABLE_MODULE
)
CostEstimator: AppFunction = (
    _get_callable(module=_COST_ESIMATOR_MODULE, callable_name=_COST_ESIMATOR_NAME)
    if _COST_ESIMATOR_NAME
    else _default_cost_estimator
)


# set up cost calculator function. This is called after job completes successfully.
# Default to a function that returns 0.
_COST_CALCULATOR_NAME = _cfg.get("cost_calculator", profile="callable_svc")
_COST_CALCULATOR_MODULE = (
    _cfg.get("cost_calculator_module", profile="callable_svc") or _CALLABLE_MODULE
)
CostCalculator: AppFunction = (
    _get_callable(module=_COST_CALCULATOR_MODULE, callable_name=_COST_CALCULATOR_NAME)
    if _COST_CALCULATOR_NAME
    else _default_cost_calculator
)
